import importlib, glob, imp, os, sys
from importlib.abc import MetaPathFinder
from importlib.util import spec_from_file_location
from importlib.machinery import ExtensionFileLoader, SourceFileLoader
from typing import List

PYX_EXT = ".pyx"
PYXDEP_EXT = ".pyxdep"
PYXBLD_EXT = ".pyxbld"


def handle_dependencies(pyxfilename):
    testing = '_test_files' in globals()
    dependfile = os.path.splitext(pyxfilename)[0] + PYXDEP_EXT

    # by default let distutils decide whether to rebuild on its own
    # (it has a better idea of what the output file will be)

    # but we know more about dependencies so force a rebuild if
    # some of the dependencies are newer than the pyxfile.
    if os.path.exists(dependfile):
        with open(dependfile) as fid:
            depends = fid.readlines()
        depends = [depend.strip() for depend in depends]

        # gather dependencies in the "files" variable
        # the dependency file is itself a dependency
        files = [dependfile]
        for depend in depends:
            fullpath = os.path.join(os.path.dirname(dependfile),
                                    depend)
            files.extend(glob.glob(fullpath))

        # if any file that the pyxfile depends upon is newer than
        # the pyx file, 'touch' the pyx file so that distutils will
        # be tricked into rebuilding it.
        for file in files:
            from distutils.dep_util import newer
            if newer(file, pyxfilename):
                print("Rebuilding %s because of %s", pyxfilename, file)
                filetime = os.path.getmtime(file)
                os.utime(pyxfilename, (filetime, filetime))


def handle_special_build(modname, pyxfilename):
    special_build = os.path.splitext(pyxfilename)[0] + PYXBLD_EXT
    ext = None
    setup_args={}
    if os.path.exists(special_build):
        # globls = {}
        # locs = {}
        # execfile(special_build, globls, locs)
        # ext = locs["make_ext"](modname, pyxfilename)
        with open(special_build) as fid:
            mod = imp.load_source("XXXX", special_build, fid)
        make_ext = getattr(mod,'make_ext',None)
        if make_ext:
            ext = make_ext(modname, pyxfilename)
            assert ext and ext.sources, "make_ext in %s did not return Extension" % special_build
        make_setup_args = getattr(mod, 'make_setup_args',None)
        if make_setup_args:
            setup_args = make_setup_args()
            assert isinstance(setup_args,dict), ("make_setup_args in %s did not return a dict"
                                         % special_build)
        assert set or setup_args, ("neither make_ext nor make_setup_args %s" % special_build)
        ext.sources = [os.path.join(os.path.dirname(special_build), source) for source in ext.sources]
    return ext, setup_args


def get_distutils_extension(modname, pyxfilename, language_level=None):
    extension_mod, setup_args = handle_special_build(modname, pyxfilename)
    if not extension_mod:
        if not isinstance(pyxfilename, str):
            # distutils is stupid in Py2 and requires exactly 'str'
            # => encode accidentally coerced unicode strings back to str
            pyxfilename = pyxfilename.encode(sys.getfilesystemencoding())
        from distutils.extension import Extension
        extension_mod = Extension(name = modname, sources=[pyxfilename])
        if language_level is not None:
            extension_mod.cython_directives = {'language_level': language_level}
    return extension_mod, setup_args


def build_module(name, pyxfilename, user_setup_args, pyxbuild_dir=None, inplace=False, language_level=None, 
                 build_in_temp=False, reload_support=True):
    assert os.path.exists(pyxfilename), "Path does not exist: %s" % pyxfilename
    handle_dependencies(pyxfilename)

    extension_mod, setup_args = get_distutils_extension(name, pyxfilename, language_level)
    build_in_temp = True
    sargs = user_setup_args.copy() if user_setup_args else dict()
    sargs.update(setup_args)
    build_in_temp = sargs.pop('build_in_temp',build_in_temp)

    from pyximport import pyxbuild
    olddir = os.getcwd()
    common = ''
    if pyxbuild_dir:
        # Windows concatenates the pyxbuild_dir to the pyxfilename when
        # compiling, and then complains that the filename is too long
        common = os.path.commonprefix([pyxbuild_dir, pyxfilename])
    if len(common) > 30:
        pyxfilename = os.path.relpath(pyxfilename)
        pyxbuild_dir = os.path.relpath(pyxbuild_dir)
        os.chdir(common)
    try:
        so_path = pyxbuild.pyx_to_dll(pyxfilename, extension_mod,
                                      force_rebuild=1,
                                      build_in_temp=build_in_temp,
                                      pyxbuild_dir=pyxbuild_dir,
                                      setup_args=sargs,
                                      inplace=inplace,
                                      reload_support=reload_support)
    finally:
        os.chdir(olddir)
    so_path = os.path.join(common, so_path)
    assert os.path.exists(so_path), "Cannot find: %s" % so_path

    junkpath = os.path.join(os.path.dirname(so_path), name+"_*")  #very dangerous with --inplace ? yes, indeed, trying to eat my files ;)
    junkstuff = glob.glob(junkpath)
    for path in junkstuff:
        if path != so_path:
            try:
                os.remove(path)
            except IOError:
                print("Couldn't remove %s", path)

    return so_path


def load_module(name, pyxfilename, pyxbuild_dir=None, is_package=False, build_inplace=False, language_level=None, so_path=None):
    try:
        if so_path is None:
            if is_package:
                module_name = name + '.__init__'
            else:
                module_name = name
            so_path = build_module(module_name, pyxfilename, pyxbuild_dir, inplace=build_inplace, language_level=language_level)
        mod = imp.load_dynamic(name, so_path)
        if is_package and not hasattr(mod, '__path__'):
            mod.__path__ = [os.path.dirname(so_path)]
        assert mod.__file__ == so_path, (mod.__file__, so_path)
    except Exception as failure_exc:
        print("Failed to load extension module: %r" % failure_exc)
        # if pyxargs.load_py_module_on_import_failure and pyxfilename.endswith('.py'):
        if False and pyxfilename.endswith('.py'):
            # try to fall back to normal import
            mod = imp.load_source(name, pyxfilename)
            assert mod.__file__ in (pyxfilename, pyxfilename+'c', pyxfilename+'o'), (mod.__file__, pyxfilename)
        else:
            tb = sys.exc_info()[2]
            import traceback
            exc = ImportError("Building module %s failed: %s" % (name, traceback.format_exception_only(*sys.exc_info()[:2])))
            if sys.version_info[0] >= 3:
                raise exc.with_traceback(tb)
            else:
                exec("raise exc, None, tb", {'exc': exc, 'tb': tb})
    return mod


class PyxImportLoader(ExtensionFileLoader):

    def __init__(self, filename, setup_args, pyxbuild_dir, inplace, language_level, reload_support):
        module_name = os.path.splitext(os.path.basename(filename))[0]
        super().__init__(module_name, filename)
        self._pyxbuild_dir = pyxbuild_dir
        self._inplace = inplace
        self._language_level = language_level
        self._setup_args = setup_args 
        self._reload_support = reload_support

    def create_module(self, spec):
        try:
            # print(f"CREATING MODULE: {spec.name} -> {spec.origin}")
            so_path = build_module(spec.name, pyxfilename=spec.origin, user_setup_args=self._setup_args, pyxbuild_dir=self._pyxbuild_dir,
                                   inplace=self._inplace, language_level=self._language_level, reload_support=self._reload_support)
            self.path = so_path
            spec.origin = so_path
            return super().create_module(spec)
        except Exception as failure_exc:
            # print("LOADING on FAILURE MODULE")
            # if pyxargs.load_py_module_on_import_failure and spec.origin.endswith('.pyx'):
            if False and spec.origin.endswith(PYX_EXT):
                spec = importlib.util.spec_from_file_location(spec.name, spec.origin,
                                                              loader=SourceFileLoader(spec.name, spec.origin))
                mod = importlib.util.module_from_spec(spec)
                assert mod.__file__ in (spec.origin, spec.origin + 'c', spec.origin + 'o'), (mod.__file__, spec.origin)
                return mod
            else:
                tb = sys.exc_info()[2]
                import traceback
                exc = ImportError("Building module %s failed: %s" % (
                    spec.name, traceback.format_exception_only(*sys.exc_info()[:2])))
                raise exc.with_traceback(tb)

    def exec_module(self, module):
        try:
            # print(f"EXEC MODULE: {module}")
            return super().exec_module(module)
        except Exception as failure_exc:
            import traceback
            print("Failed to load extension module: %r" % failure_exc)
            raise ImportError("Executing module %s failed %s" % (
                    module.__file__, traceback.format_exception_only(*sys.exc_info()[:2])))


class CustomPyxImportMetaFinder(MetaPathFinder):

    def __init__(self, modules_to_check: List[str], extension=PYX_EXT, setup_args=None, pyxbuild_dir=None, inplace=False, language_level=None, reload_support=True):
        self.valid_modules = modules_to_check
        self.pyxbuild_dir = pyxbuild_dir
        self.inplace = inplace
        self.language_level = language_level
        self.extension = extension
        self.setup_args = setup_args if setup_args else dict()
        self.reload_support = reload_support

    def find_spec(self, fullname, path, target=None):
        def _is_valid(module):
            if not self.valid_modules:
                return True
            for m in self.valid_modules:
                if module.startswith(m):
                    return True
            return False

        if not path:
            path = [os.getcwd()]  # top level import --
        if "." in fullname:
            *parents, name = fullname.split(".")
        else:
            name = fullname
        for entry in path:
            if os.path.isdir(os.path.join(entry, name)):
                # this module has child modules
                filename = os.path.join(entry, name, "__init__" + self.extension)
                submodule_locations = [os.path.join(entry, name)]
            else:
                filename = os.path.join(entry, name + self.extension)
                submodule_locations = None
            if not os.path.exists(filename):
                continue

            if not _is_valid(fullname):
                continue

            return spec_from_file_location(
                fullname, filename,
                loader=PyxImportLoader(filename, self.setup_args, self.pyxbuild_dir, self.inplace, self.language_level, self.reload_support),
                submodule_search_locations=submodule_locations)

        return None  # we don't know how to import this


__pyx_finder_installed = False

def pyx_install_loader(modules_to_check: List[str]):
    import numpy as np
    import pyximport
    global __pyx_finder_installed

    if not __pyx_finder_installed:
        build_dir = os.path.expanduser("~/.pyxbld")
        setup_args = {'include_dirs': np.get_include()}
        sys.meta_path.insert(0, CustomPyxImportMetaFinder(
            modules_to_check,
            PYX_EXT, setup_args=setup_args, pyxbuild_dir=build_dir, 
            language_level=3, reload_support=True
        ))
        pyximport.install(setup_args=setup_args, build_dir=build_dir, reload_support=True, language_level=3)




import datetime
import itertools
import os
import platform
import shutil
import subprocess
import sysconfig

from pathlib import Path
import os
import numpy as np
import toml
from Cython.Build import build_ext
from Cython.Build import cythonize
from Cython.Compiler import Options
from Cython.Compiler.Version import version as cython_compiler_version
from setuptools import Distribution
from setuptools import Extension


BUILD_MODE = os.getenv("BUILD_MODE", "release")
PROFILE_MODE = bool(os.getenv("PROFILE_MODE", ""))
ANNOTATION_MODE = bool(os.getenv("ANNOTATION_MODE", ""))
BUILD_DIR = "build/optimized"
COPY_TO_SOURCE = os.getenv("COPY_TO_SOURCE", "true") == "true"

################################################################################
#  CYTHON BUILD
################################################################################
# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html

Options.docstrings = True  # Include docstrings in modules
Options.fast_fail = True  # Abort the compilation on the first error occurred
Options.annotate = ANNOTATION_MODE  # Create annotated HTML files for each .pyx
if ANNOTATION_MODE:
    Options.annotate_coverage_xml = "coverage.xml"
Options.fast_fail = True  # Abort compilation on first error
Options.warning_errors = True  # Treat compiler warnings as errors
Options.extra_warnings = True

CYTHON_COMPILER_DIRECTIVES = {
    "language_level": "3",
    "cdivision": True,  # If division is as per C with no check for zero (35% speed up)
    "nonecheck": True,  # Insert extra check for field access on C extensions
    "embedsignature": True,  # If docstrings should be embedded into C signatures
    "profile": PROFILE_MODE,  # If we're debugging or profiling
    "linetrace": PROFILE_MODE,  # If we're debugging or profiling
    "warn.maybe_uninitialized": True,
}


def _build_extensions() -> list[Extension]:
    # Regarding the compiler warning: #warning "Using deprecated NumPy API,
    # disable it with " "#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION"
    # https://stackoverflow.com/questions/52749662/using-deprecated-numpy-api
    # From the Cython docs: "For the time being, it is just a warning that you can ignore."
    define_macros: list[tuple[str, str | None]] = [
        ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"),
    ]
    if PROFILE_MODE or ANNOTATION_MODE:
        # Profiling requires special macro directives
        define_macros.append(("CYTHON_TRACE", "1"))

    extra_compile_args = []
    extra_link_args = []
    # extra_link_args = RUST_LIBS

    if platform.system() != "Windows":
        # Suppress warnings produced by Cython boilerplate
        extra_compile_args.append("-Wno-unreachable-code")
        if BUILD_MODE == "release":
            extra_compile_args.append("-O2")
            extra_compile_args.append("-pipe")

    if platform.system() == "Windows":
        extra_link_args += [
            "AdvAPI32.Lib",
            "bcrypt.lib",
            "Crypt32.lib",
            "Iphlpapi.lib",
            "Kernel32.lib",
            "ncrypt.lib",
            "Netapi32.lib",
            "ntdll.lib",
            "Ole32.lib",
            "OleAut32.lib",
            "Pdh.lib",
            "PowrProf.lib",
            "Psapi.lib",
            "schannel.lib",
            "secur32.lib",
            "Shell32.lib",
            "User32.Lib",
            "UserEnv.Lib",
            "WS2_32.Lib",
        ]

    print("Creating C extension modules...")
    print(f"define_macros={define_macros}")
    print(f"extra_compile_args={extra_compile_args}")

    return [
        Extension(
            name=str(pyx.relative_to(".")).replace(os.path.sep, ".")[:-4],
            sources=[str(pyx)],
            include_dirs=[np.get_include()], #, *RUST_INCLUDES],
            define_macros=define_macros,
            language="c",
            extra_link_args=extra_link_args,
            extra_compile_args=extra_compile_args,
        ) for pyx in itertools.chain(Path("src/qubx").rglob("*.pyx"))
    ]


def _build_distribution(extensions: list[Extension]) -> Distribution:
    nthreads = os.cpu_count() or 1
    if platform.system() == "Windows":
        nthreads = min(nthreads, 60)
    print(f"nthreads={nthreads}")

    distribution = Distribution(
        {
            "name": "qubx",
            "ext_modules": cythonize(
                module_list=extensions,
                compiler_directives=CYTHON_COMPILER_DIRECTIVES,
                nthreads=nthreads,
                build_dir=BUILD_DIR,
                gdb_debug=PROFILE_MODE,
            ),
            "zip_safe": False,
        },
    )
    return distribution


def _copy_build_dir_to_project(cmd: build_ext) -> None:
    # Copy built extensions back to the project tree
    for output in cmd.get_outputs():
        relative_extension = Path(output).relative_to(cmd.build_lib)
        if not Path(output).exists():
            continue

        # Copy the file and set permissions
        shutil.copyfile(output, relative_extension)
        mode = relative_extension.stat().st_mode
        mode |= (mode & 0o444) >> 2
        relative_extension.chmod(mode)

    print("Copied all compiled dynamic library files into source")


def _strip_unneeded_symbols() -> None:
    try:
        print("Stripping unneeded symbols from binaries...")
        for so in itertools.chain(Path("nautilus_trader").rglob("*.so")):
            if platform.system() == "Linux":
                strip_cmd = ["strip", "--strip-unneeded", so]
            elif platform.system() == "Darwin":
                strip_cmd = ["strip", "-x", so]
            else:
                raise RuntimeError(f"Cannot strip symbols for platform {platform.system()}")
            subprocess.run(
                strip_cmd,  # type: ignore [arg-type] # noqa
                check=True,
                capture_output=True,
            )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Error when stripping symbols.\n{e}") from e


def build() -> None:
    """
    Construct the extensions and distribution.
    """
    # _build_rust_libs()
    # _copy_rust_dylibs_to_project()

    if True: #not PYO3_ONLY:
        # Create C Extensions to feed into cythonize()
        extensions = _build_extensions()
        distribution = _build_distribution(extensions)

        # Build and run the command
        print("Compiling C extension modules...")
        cmd: build_ext = build_ext(distribution)
        # if PARALLEL_BUILD:
            # cmd.parallel = os.cpu_count()
        cmd.ensure_finalized()
        cmd.run()

        if COPY_TO_SOURCE:
            # Copy the build back into the source tree for development and wheel packaging
            _copy_build_dir_to_project(cmd)

    if BUILD_MODE == "release" and platform.system() in ("Linux", "Darwin"):
        # Only strip symbols for release builds
        _strip_unneeded_symbols()


if __name__ == "__main__":
    qubx_platform = toml.load("pyproject.toml")["tool"]["poetry"]["version"]
    print("\033[36m")
    print("=====================================================================")
    print(f"Qubx Builder {qubx_platform}")
    print("=====================================================================\033[0m")
    print(f"System: {platform.system()} {platform.machine()}")
    # print(f"Clang:  {_get_clang_version()}")
    # print(f"Rust:   {_get_rustc_version()}")
    print(f"Python: {platform.python_version()}")
    print(f"Cython: {cython_compiler_version}")
    print(f"NumPy:  {np.__version__}\n")

    print(f"BUILD_MODE={BUILD_MODE}")
    print(f"BUILD_DIR={BUILD_DIR}")
    print(f"PROFILE_MODE={PROFILE_MODE}")
    print(f"ANNOTATION_MODE={ANNOTATION_MODE}")
    # print(f"PARALLEL_BUILD={PARALLEL_BUILD}")
    print(f"COPY_TO_SOURCE={COPY_TO_SOURCE}")
    # print(f"PYO3_ONLY={PYO3_ONLY}\n")

    print("Starting build...")
    ts_start = datetime.datetime.now(datetime.timezone.utc)
    build()
    print(f"Build time: {datetime.datetime.now(datetime.timezone.utc) - ts_start}")
    print("\033[32m" + "Build completed" + "\033[0m")

# # See if Cython is installed
# try:
#     from Cython.Build import cythonize
# # Do nothing if Cython is not available
# except ImportError:
#     # Got to provide this function. Otherwise, poetry will fail
#     def build(setup_kwargs):
#         pass

# # Cython is installed. Compile
# else:
#     from setuptools import Extension
#     from setuptools.dist import Distribution
#     from distutils.command.build_ext import build_ext

#     # This function will be executed in setup.py:
#     def build(setup_kwargs):
#         # The file you want to compile
#         extensions = [
#             "src/qubx/core/series.pyx",
#             "src/qubx/core/utils.pyx",
#             "src/qubx/ta/indicators.pyx",
#         ]

#         # gcc arguments hack: enable optimizations
#         os.environ['CFLAGS'] = '-O3'

#         # Build
#         import numpy as np
#         setup_kwargs.update({
#             'ext_modules': cythonize(
#                 extensions,
#                 language_level=3,
#                 compiler_directives={'linetrace': True},
#                 include_path=[np.get_include()]
#             ),
#             'cmdclass': {'build_ext': build_ext}
#         })


import numpy as np
import pandas as pd
from os.path import exists, join, split, basename
# from dateutil import parser
from tqdm.notebook import tqdm
from typing import Any, Callable, Dict, List
from binance.client import Client, HistoricalKlinesType, BinanceAPIException
import requests

from qubx import logger
from qubx.utils.misc import makedirs, get_local_qubx_folder

DEFALT_LOCAL_FILE_STORAGE = makedirs(get_local_qubx_folder(), 'data/export/binance_trades/')

_DEFAULT_MARKET_DATA_DB = 'md'
BINANCE_DATA_STORAGE = "https://s3-ap-northeast-1.amazonaws.com"
BINANCE_DATA_URL = "https://data.binance.vision/"

    
def get_binance_symbol_info_for_type(market_types: List[str]) -> Dict[str, Dict[str, Any]]:
    """
    Get list of all symbols from binance for given list of market types:
    possible types are: SPOT, FUTURES, COINSFUTURES
    
    >>> get_binance_symbol_info_for_type('FUTURES')
    
    :param market_type: SPOT, FUTURES (UM) or COINSFUTURES (CM)
    """
    client = Client()
    infos = {}
    for market_type in (market_types if not isinstance(market_types, str) else [market_types]):
        if market_type in ['FUTURES', 'UM']:
            infos['binance.um'] = client.futures_exchange_info()

        elif market_type in ['COINSFUTURES', 'CM']:
            infos['binance.cm'] = client.futures_coin_exchange_info()

        elif market_type == 'SPOT':
            infos['binance'] = client.get_exchange_info()
        else:
            raise ValueError("Only 'FUTURES | UM', 'COINSFUTURES | CM' or 'SPOT' are supported for market_type")

    return infos

            
def fetch_file(url, local_file_storage, chunk_size=1024*1024):
    """
    Load file from url and store it to specified storage
    """
    file = split(url)[-1]
    response = requests.get(url, stream=True)
    with open(join(local_file_storage, file), "wb") as handle:
        for data in tqdm(response.iter_content(chunk_size=chunk_size)):
            handle.write(data)


def get_trades_files(symbol: str, instr_type: str, instr_subtype: str):
    """
    Get list of trades files for specified instrument from Binance datastorage
    """
    if instr_type.lower() == 'spot':
        instr_subtype = ''
    filter_str =  join("data", instr_type.lower(), instr_subtype.lower(), "monthly", "trades", symbol.upper())
    pg = requests.get(f"{BINANCE_DATA_STORAGE}/data.binance.vision?prefix={filter_str}/")
    info = pd.read_xml(pg.text)
    return [k for k in info.Key.dropna() if k.endswith('.zip')]


def load_trades_for(symbol, instr_type='futures', instr_subtype='um', local_file_storage=DEFALT_LOCAL_FILE_STORAGE):
    """
    Load trades from Binance data storage
    >>> load_trades_for('ETHUSDT', 'futures', 'um')
    """
    local_file_storage = makedirs(local_file_storage)

    f_list = get_trades_files(symbol, instr_type, instr_subtype)
    for r_file in tqdm(f_list):
        dest_dir = join(local_file_storage, symbol.upper())
        dest_file = join(dest_dir, basename(r_file))
        if not exists(dest_file):
            fetch_file(join(BINANCE_DATA_URL, r_file), dest_dir)
        else:
            logger.info(f"{dest_file} already loaded, skipping ...")


# class BinanceHist:
#     START_SPOT_HIST = pd.Timestamp('2017-01-01')
#     START_FUT_HIST = pd.Timestamp('2019-09-08')
    
#     def __init__(self, api_key, secret_key):
#         self.client = Client(api_key=api_key, api_secret=secret_key)
    
#     @staticmethod
#     def from_env(path):
#         if exists(path):
#             return BinanceHist(**get_env_data_as_dict(path))
#         else:
#             raise ValueError(f"Can't find env file at {path}")
    
#     @staticmethod
#     def _t2s(time):
#         _t = pd.Timestamp(time) if not isinstance(time, pd.Timestamp) else time
#         return _t.strftime('%d %b %Y %H:%M:%S')
    
#     def load_hist_spot_data(self, symbol, timeframe, start: pd.Timestamp, stop: pd.Timestamp, step='4W', timeout_sec=2):
#         return self.load_hist_data(symbol, timeframe, start, stop, HistoricalKlinesType.SPOT, step, timeout_sec)
    
#     def load_hist_futures_data(self, symbol, timeframe, start: pd.Timestamp, stop: pd.Timestamp, step='4W', timeout_sec=2):
#         return self.load_hist_data(symbol, timeframe, start, stop, HistoricalKlinesType.FUTURES, step, timeout_sec)

#     def update_hist_data(self, symbol, stype, timeframe, exchange_id, drop_prev=False, step='4W', timeout_sec=2):
#         """
#         Update loaded data in z db:
        
#         >>> bh = BinanceHist.from_env('.env')
#         >>> bh.update_hist_data('BTCUSDT', 'FUTURES', '1Min', 'BINANCEF')
#         >>> bh.update_hist_data('BTCUSDT', 'SPOT', '1Min', 'BINANCE')
#         """
#         if isinstance(symbol, (tuple, list)):
#             for s in symbol:
#                 print(f"Working on {symbol.index(s)}th from {len(symbol)} symbols...")
#                 self.update_hist_data(s, stype, timeframe, exchange_id, drop_prev=drop_prev, step=step, timeout_sec=timeout_sec)
#             return
                
#         klt = {
#             'futures': [HistoricalKlinesType.FUTURES, BinanceHist.START_FUT_HIST],
#             'spot':    [HistoricalKlinesType.SPOT, BinanceHist.START_SPOT_HIST]
#         }.get(stype.lower()) 
        
#         if klt is None:
#             raise ValueError(f"Unknown instrument type '{stype}' !")
            
#         tD = pd.Timedelta(timeframe)
#         now = self._t2s(pd.Timestamp(datetime.datetime.now(pytz.UTC).replace(second=0)) - tD)
        
#         path = f'm1/{exchange_id}:{symbol}'
#         if drop_prev:
#             print(f' > Deleting data at {path} ...')
#             z_del(path, dbname=_DEFAULT_MARKET_DATA_DB)
        
#         mc = MongoController(dbname=_DEFAULT_MARKET_DATA_DB)
        
#         hist = z_ld(path, dbname=_DEFAULT_MARKET_DATA_DB)
#         if hist is None:
#             last_known_date = klt[1]
#         else:   
#             last_known_date = hist.index[-1]
            
#             # 're-serialize' data to be accessible through gridfs
#             if mc.get_count(path) == 0:
#                 z_save(path, hist, is_serialize=False, dbname=_DEFAULT_MARKET_DATA_DB)

#         s_time = last_known_date.round(tD)
            
#         # pull up recent data
#         def __cb(data):
#             # print(">>>>>> new data: ", len(data))
#             mc.append_data(path, data)
            
#         self.load_hist_data(symbol, timeframe, s_time, now, klt[0], step=step, timeout_sec=timeout_sec, new_data_callback=__cb)
#         mc.close()
        
#         # drop _id (after update through MongoController)
#         pdata = z_ld(path, dbname=_DEFAULT_MARKET_DATA_DB)
#         if pdata is not None:
#             pdata = pdata.drop(columns='_id')
#         z_save(path, pdata, dbname=_DEFAULT_MARKET_DATA_DB)
    
#     def load_hist_data(self, symbol, timeframe, start: pd.Timestamp, stop: pd.Timestamp, ktype, step='4W', timeout_sec=2,
#                        new_data_callback: Callable[[pd.DataFrame], None]=None):
#         start = pd.Timestamp(start) if not isinstance(start, pd.Timestamp) else start
#         stop = pd.Timestamp(stop) if not isinstance(stop, pd.Timestamp) else stop
#         df = pd.DataFrame()
        
#         _tf_str = timeframe[:2].lower()
#         tD = pd.Timedelta(timeframe)
#         now = self._t2s(stop)
#         tlr = pd.DatetimeIndex([start]).append(pd.date_range(start, now, freq=step).append(pd.DatetimeIndex([now])))
        
#         print(f' >> Loading {green(symbol)} {yellow(timeframe)} for [{red(start)}  -> {red(now)}]')
#         s = tlr[0]
#         for e in tqdm(tlr[1:]):
#             if s + tD < e:
#                 _start, _stop = self._t2s(s + tD), self._t2s(e)
#                 # print(_start, _stop)
#                 chunk = None
#                 nerr = 0
#                 while nerr < 3:
#                     try:
#                         chunk = self.client.get_historical_klines(symbol, _tf_str, _start, _stop, klines_type=ktype, limit=1000)
#                         nerr = 100
#                     except BinanceAPIException as a:
#                         print('.', end='')
#                         break
#                     except Exception as err:
#                         nerr +=1
#                         print(red(str(err)))
#                         time.sleep(10)

#                 if chunk:
#                     data = pd.DataFrame(chunk, columns = ['timestamp', 'open', 'high', 'low', 'close', 'volume', 'close_time', 'quote_av', 'trades', 'tb_base_av', 'tb_quote_av', 'ignore' ])
#                     data.index = pd.to_datetime(data['timestamp'].rename('time'), unit='ms')
#                     data = data.drop(columns=['timestamp', 'close_time', 'quote_av', 'trades', 'tb_base_av', 'tb_quote_av', 'ignore' ]).astype(float)
#                     df = df.append(data)
#                     # callback on new data
#                     if new_data_callback is not None:
#                         new_data_callback(data)
                        
#                 s = e
#                 time.sleep(timeout_sec)
#         return df
    


from datetime import datetime
from typing import List, Optional, Union
import numpy as np
import re

import pandas as pd

UNIX_T0 = np.datetime64('1970-01-01T00:00:00')


time_to_str = lambda t, u='us': np.datetime_as_string(t if isinstance(t, np.datetime64) else np.datetime64(t, u), unit=u)


def convert_tf_str_td64(c_tf: str) -> np.timedelta64:
    """
    Convert string timeframe to timedelta64

    '15Min' -> timedelta64(15, 'm') etc
    """
    _t = re.findall('(\d+)([A-Za-z]+)', c_tf)
    _dt = 0
    for g in _t:
        unit = g[1].lower()
        n = int(g[0])
        u1 = unit[0]
        u2 = unit[:2]
        unit = u1

        if u1 in ['d', 'w']:
            unit = u1.upper()

        if u1 in ['y']:
            n = 356 * n
            unit = 'D'

        if u2 in ['ms', 'ns', 'us', 'ps']:
            unit = u2

        _dt += np.timedelta64(n, unit)
    
    return _dt


def convert_seconds_to_str(seconds: int) -> str:
    """
    Convert seconds to string representation: 310 -> '5Min10S' etc
    """
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    r = ''
    if days > 0:
        r += '%dD' % days
    if hours > 0:
        r += '%dH' % hours
    if minutes > 0:
        r += '%dMin' % minutes
    if seconds > 0:
        r += '%dS' % seconds
    return r


def floor_t64(time: Union[np.datetime64, datetime], dt: Union[np.timedelta64, int, str]):
    """
    Floor timestamp by dt
    """
    if isinstance(dt, int):
        dt = np.timedelta64(dt, 's')

    if isinstance(dt, str):
        dt = convert_tf_str_td64(dt)

    if isinstance(time, datetime):
        time = np.datetime64(time)

    return time - (time - UNIX_T0) % dt


def infer_series_frequency(series: Union[List, pd.DataFrame, pd.Series, pd.DatetimeIndex]) -> np.timedelta64:
    """
    Infer frequency of given timeseries

    :param series: Series, DataFrame, DatetimeIndex or list of timestamps object
    :return: timedelta for found frequency
    """
    if isinstance(series, (pd.DataFrame, pd.Series, pd.DatetimeIndex)):
        times_index = (series if isinstance(series, pd.DatetimeIndex) else series.index).to_pydatetime()
    elif isinstance(series, (set, list, tuple)):
        times_index = np.array(series)
    elif isinstance(series, np.ndarray):
        times_index = series
    else:
        raise ValueError("Can't recognize input data")

    if times_index.shape[0] < 2:
        raise ValueError("Series must have at least 2 points to determ frequency")

    values = np.array(sorted([(x if isinstance(x, np.timedelta64) else x.total_seconds()) for x in np.abs(np.diff(times_index))]))
    diff = np.concatenate(([1], np.diff(values)))
    idx = np.concatenate((np.where(diff)[0], [len(values)]))
    freqs = dict(zip(values[idx[:-1]], np.diff(idx)))
    return np.timedelta64(max(freqs, key=freqs.get))


def handle_start_stop(s: Optional[str], e: Optional[str], convert=str) -> tuple:
    """
    Process start/stop times

        handle_start_stop('2020-01-01', '2020-02-01') # 2020-01-01, 2020-02-01
        handle_start_stop('2020-02-01', '2020-01-01') # 2020-01-01, 2020-02-01
        handle_start_stop('2020-01-01', '1w')         # 2020-01-01, 2020-01-01 + 1week
        handle_start_stop('1w', '2020-01-01')         # 2020-01-01 - 1week, '2020-01-01'
        handle_start_stop('2020-01-01', '-1w')        # 2020-01-01 - 1week, 2020-01-01,
        handle_start_stop(None, '2020-01-01')         # None, '2020-01-01'
        handle_start_stop('2020-01-01', None)         # '2020-01-01', None 
        handle_start_stop(None, None)                 # None, None 

    """
    def _h_time_like(x):
        try: 
            return pd.Timestamp(x), False
        except:
            try: return pd.Timedelta(x), True
            except: pass
        return None, None

    t0, d0 = _h_time_like(s) if s else (None, False)
    t1, d1 = _h_time_like(e) if e else (None, False)
    converts = lambda xs: [convert(xs[0]) if xs[0] else None, convert(xs[1]) if xs[1] else None]

    if not t1 and not t0: return None, None
    
    if d0 and d1: raise ValueError("Start and stop can't both be deltas !")

    if d0:
        if not t1: raise ValueError("First argument is delta but stop time is not defined !")
        return converts(sorted([t1 - abs(t0), t1]))
    if d1:
        if not t0: raise ValueError("Second argument is delta but start time is not defined !")
        return converts(sorted([t0, t0 + t1]))

    if t0 and t1:
        return converts(sorted([t0, t1]))

    return converts([t0, t1])


from typing import List, Union, Optional, Iterable, Any
from os.path import exists
import numpy as np
import pandas as pd
import pyarrow as pa
from pyarrow import csv

from qubx import logger
from qubx.core.series import TimeSeries, OHLCV, time_as_nsec, Quote, Trade
from qubx.utils.time import infer_series_frequency, handle_start_stop

_DT = lambda x: pd.Timedelta(x).to_numpy().item()
D1, H1 = _DT('1D'), _DT('1H')

DEFAULT_DAILY_SESSION = (_DT('00:00:00.100'), _DT('23:59:59.900'))
STOCK_DAILY_SESSION = (_DT('9:30:00.100'), _DT('15:59:59.900'))
CME_FUTURES_DAILY_SESSION = (_DT('8:30:00.100'), _DT('15:14:59.900'))


def _recognize_t(t: Union[int, str], defaultvalue, timeunit) -> int:
    if isinstance(t, (str, pd.Timestamp)):
        try:
            return np.datetime64(t, timeunit)
        except:
            pass
    return defaultvalue


def _find_column_index_in_list(xs, *args):
    xs = [x.lower() for x in xs]
    for a in args:
        ai = a.lower()
        if ai in xs:
            return xs.index(ai)
    raise IndexError(f"Can't find any from {args} in list: {xs}")


class DataProcessor:
    """
    Common interface for data processor with default aggregating into list implementation
    """
    def __init__(self) -> None:
        self.buffer = {}
        self._column_names = []

    def start_processing(self, column_names: List[str]):
        self._column_names = column_names
        self.buffer = {c: [] for c in column_names}

    def process_data(self, columns_data: list) -> Optional[Iterable]:
        for i, c in enumerate(columns_data):
            self.buffer[self._column_names[i]].append(c)
        return None

    def get_result(self) -> Any:
        return self.buffer


class DataReader:
    """
    Common interface for data reader
    """
    _processor: DataProcessor

    def __init__(self, processor=None) -> None:
        self._processor = DataProcessor() if processor is None else processor

    def read(self, start: Optional[str]=None, stop: Optional[str]=None) -> Any:
        pass

    
class QuotesDataProcessor(DataProcessor):
    """
    Process quotes data and collect them as list
    """
    def start_processing(self, fieldnames: List[str]):
        self.buffer = list()
        self._time_idx = _find_column_index_in_list(fieldnames, 'time', 'timestamp', 'datetime')
        self._bid_idx = _find_column_index_in_list(fieldnames, 'bid')
        self._ask_idx = _find_column_index_in_list(fieldnames, 'ask')
        self._bidvol_idx = _find_column_index_in_list(fieldnames, 'bidvol', 'bid_vol', 'bidsize', 'bid_size')
        self._askvol_idx = _find_column_index_in_list(fieldnames, 'askvol', 'ask_vol', 'asksize', 'ask_size')

    def process_data(self, columns_data: list) -> Optional[Iterable]:
        tms = columns_data[self._time_idx] 
        bids = columns_data[self._bid_idx]
        asks = columns_data[self._ask_idx]
        bidvol = columns_data[self._bidvol_idx]
        askvol = columns_data[self._askvol_idx]
        for i in range(len(tms)):
            self.buffer.append(Quote(tms[i], bids[i], asks[i], bidvol[i], askvol[i]))
        return None


class QuotesFromOHLCVDataProcessor(DataProcessor):
    """
    Process OHLC and generate Quotes (+ Trades) from it
    """
    def __init__(self, trades: bool=False, 
                 default_bid_size=1e9,  # default bid/ask is big
                 default_ask_size=1e9,  # default bid/ask is big
                 daily_session_start_end=DEFAULT_DAILY_SESSION,
                 spread=0.0,
                ) -> None:
        super().__init__()
        self._trades = trades
        self._bid_size = default_bid_size
        self._ask_size = default_ask_size
        self._s2 = spread / 2.0
        self._d_session_start = daily_session_start_end[0]
        self._d_session_end = daily_session_start_end[1]

    def start_processing(self, fieldnames: List[str]):
        self._time_idx = _find_column_index_in_list(fieldnames, 'time', 'timestamp', 'datetime', 'date')
        self._open_idx = _find_column_index_in_list(fieldnames, 'open')
        self._high_idx = _find_column_index_in_list(fieldnames, 'high')
        self._low_idx = _find_column_index_in_list(fieldnames, 'low')
        self._close_idx = _find_column_index_in_list(fieldnames, 'close')
        self._volume_idx = None
        self._timeframe = None

        try:
            self._volume_idx = _find_column_index_in_list(fieldnames, 'volume', 'vol')
        except:
            pass

        self.buffer = []

    def process_data(self, data: list) -> Optional[Iterable]:
        s2 = self._s2
        if self._timeframe is None:
            _freq = infer_series_frequency(data[self._time_idx])
            self._timeframe = _freq.astype('timedelta64[s]')

            # - timestamps when we emit simulated quotes
            dt = _freq.astype('timedelta64[ns]').item()
            if dt < D1:
                self._t_start = dt // 10
                self._t_mid1 = dt // 2 - dt // 10
                self._t_mid2 = dt // 2 + dt // 10
                self._t_end = dt - dt // 10
            else:
                self._t_start = self._d_session_start
                self._t_mid1 = dt // 2 - H1
                self._t_mid2 = dt // 2 + H1
                self._t_end = self._d_session_end

        # - input data
        times = data[self._time_idx]
        opens = data[self._open_idx]
        highs = data[self._high_idx]
        lows = data[self._low_idx]
        closes = data[self._close_idx]
        volumes = data[self._volume_idx] if self._volume_idx else None
        if volumes is None and self._trades:
            logger.warning("Input OHLC data doesn't contain volume information so trades can't be emulated !")
            self._trades = False

        for i in range(len(times)):
            ti, o, h, l, c = times[i].astype('datetime64[ns]'), opens[i], highs[i], lows[i], closes[i]

            if self._trades:
                rv = volumes[i] / (h - l) 

            # - opening quote
            self.buffer.append(Quote(ti + self._t_start, o - s2, o + s2, self._bid_size, self._ask_size))

            if c >= o:
                if self._trades:
                    self.buffer.append(Trade(ti + self._t_start, o - s2, rv * (o - l))) # sell 1
                self.buffer.append(Quote(ti + self._t_mid1, l - s2, l + s2, self._bid_size, self._ask_size))

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid1, l + s2, rv * (c - o)))  # buy 1
                self.buffer.append(Quote(ti + self._t_mid2, h - s2, h + s2, self._bid_size, self._ask_size))

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid2, h - s2, rv * (h - c)))  # sell 2
            else:
                if self._trades:
                    self.buffer.append(Trade(ti + self._t_start, o + s2, rv * (h - o))) # buy 1
                self.buffer.append(Quote(ti + self._t_mid1, h - s2, h + s2, self._bid_size, self._ask_size))

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid1, h - s2, rv * (o - c))) # sell 1
                self.buffer.append(Quote(ti + self._t_mid2, l - s2, l + s2, self._bid_size, self._ask_size))

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid2, l + s2, rv * (c - l))) # buy 2

            # - closing quote
            self.buffer.append(Quote(ti + self._t_end, c - s2, c + s2, self._bid_size, self._ask_size))

        return None

    def get_result(self) -> Any:
        return self.buffer


class OhlcvDataProcessor(DataProcessor):
    """
    Process data and convert it to Qube OHLCV timeseries 
    """
    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = name

    def start_processing(self, fieldnames: List[str]):
        self._time_idx = _find_column_index_in_list(fieldnames, 'time', 'timestamp', 'datetime', 'date')
        self._open_idx = _find_column_index_in_list(fieldnames, 'open')
        self._high_idx = _find_column_index_in_list(fieldnames, 'high')
        self._low_idx = _find_column_index_in_list(fieldnames, 'low')
        self._close_idx = _find_column_index_in_list(fieldnames, 'close')
        self._volume_idx = None
        self._timeframe = None

        try:
            self._volume_idx = _find_column_index_in_list(fieldnames, 'volume', 'vol')
        except:
            pass

        self.ohlc = None

    def process_data(self, data: list) -> Optional[Iterable]:
        if self._timeframe is None:
            self._timeframe = infer_series_frequency(data[self._time_idx]).astype('timedelta64[s]')

            # - create instance after first data received
            self.ohlc = OHLCV(self._name, self._timeframe)

        self.ohlc.append_data(
            data[self._time_idx],
            data[self._open_idx], data[self._high_idx], data[self._low_idx], data[self._close_idx], 
            data[self._volume_idx] if self._volume_idx else []
        )
        return None

    def get_result(self) -> Any:
        return self.ohlc


class OhlcvPandasDataProcessor(DataProcessor):
    """
    Process data and convert it to pandas OHLCV dataframes 
    """
    def __init__(self) -> None:
        super().__init__()

    def start_processing(self, fieldnames: List[str]):
        self._time_idx = _find_column_index_in_list(fieldnames, 'time', 'timestamp', 'datetime', 'date')
        self._open_idx = _find_column_index_in_list(fieldnames, 'open')
        self._high_idx = _find_column_index_in_list(fieldnames, 'high')
        self._low_idx = _find_column_index_in_list(fieldnames, 'low')
        self._close_idx = _find_column_index_in_list(fieldnames, 'close')
        self._volume_idx = None
        self._timeframe = None

        try:
            self._volume_idx = _find_column_index_in_list(fieldnames, 'volume', 'vol')
        except:
            pass

        # self.ohlc = pd.DataFrame()

        self._time = np.array([], dtype=np.datetime64)
        self._open = np.array([])
        self._high = np.array([])
        self._low = np.array([])
        self._close = np.array([])
        self._volume = np.array([])

    def process_data(self, data: list) -> Optional[Iterable]:
        # p = pd.DataFrame({
        #     'open': data[self._open_idx], 
        #     'high': data[self._high_idx], 
        #     'low': data[self._low_idx], 
        #     'close': data[self._close_idx], 
        #     'volume': data[self._volume_idx] if self._volume_idx else []},
        #     index = data[self._time_idx]
        # )
        # self.ohlc = pd.concat((self.ohlc, p), axis=0, sort=True, copy=True)
        self._time = np.concatenate((self._time, data[self._time_idx]))
        self._open = np.concatenate((self._open, data[self._open_idx]))
        self._high = np.concatenate((self._high, data[self._high_idx]))
        self._low = np.concatenate((self._low, data[self._low_idx]))
        self._close = np.concatenate((self._close, data[self._close_idx]))
        if self._volume_idx:
            self._volume = np.concatenate((self._volume, data[self._volume_idx]))

        return None

    def get_result(self) -> Any:
        # self.ohlc.index.name = 'time'
        # return self.ohlc

        return pd.DataFrame(
            {
                'open': self._open, 
                'high': self._high, 
                'low': self._low, 
                'close': self._close, 
                'volume': self._volume if self._volume_idx else []
            },
            index = self._time
        ).sort_index()
 

class CsvDataReader(DataReader):
    """
    CSV data file reader
    """

    def __init__(self, path: str, processor: DataProcessor=None, timestamp_parsers=None) -> None:
        if not exists(path):
            raise ValueError(f"CSV file not found at {path}")
        super().__init__(processor)
        self.time_parsers = timestamp_parsers
        self.path = path

    def __find_time_idx(self, arr: pa.ChunkedArray, v) -> int:
        ix = arr.index(v).as_py()
        if ix < 0:
            for c in arr.iterchunks():
                a = c.to_numpy()
                ix = np.searchsorted(a, v, side='right')
                if ix > 0 and ix < len(c):
                    ix = arr.index(a[ix]).as_py() - 1
                    break
        return ix

    def read(self, start: Optional[str]=None, stop: Optional[str]=None) -> Any:
        convert_options = None
        if self.time_parsers:
            convert_options=csv.ConvertOptions(timestamp_parsers=self.time_parsers)

        table = csv.read_csv(
            self.path, 
            parse_options=csv.ParseOptions(ignore_empty_lines=True),
            convert_options=convert_options
        )
        fieldnames =  table.column_names  

        # - try to find range to load  
        start_idx, stop_idx = 0, table.num_rows
        try:
            _time_field_idx = _find_column_index_in_list(fieldnames, 'time', 'timestamp', 'datetime', 'date')
            _time_type = table.field(_time_field_idx).type
            _time_unit = _time_type.unit if hasattr(_time_type, 'unit') else 's'
            _time_data = table[_time_field_idx]

            # - check if need convert time to primitive types (i.e. Date32 -> timestamp[x])
            _time_cast_function = lambda xs: xs
            if _time_type != pa.timestamp(_time_unit):
                _time_cast_function = lambda xs: xs.cast(pa.timestamp(_time_unit)) 
                _time_data = _time_cast_function(_time_data)

            # - preprocessing start and stop
            t_0, t_1 = handle_start_stop(start, stop, convert=lambda x: _recognize_t(x, None, _time_unit))

            # - check requested range
            if t_0:
                start_idx = self.__find_time_idx(_time_data, t_0)
                if start_idx >= table.num_rows:
                    # no data for requested start date
                    return None

            if t_1:
                stop_idx = self.__find_time_idx(_time_data, t_1)
                if stop_idx < 0 or stop_idx < start_idx:
                    stop_idx = table.num_rows

        except Exception as exc:
            logger.warning(exc)
            logger.info('loading whole file')

        length = (stop_idx - start_idx + 1)
        self._processor.start_processing(fieldnames)
        selected_table = table.slice(start_idx, length)
        n_chunks = selected_table[table.column_names[0]].num_chunks
        for n in range(n_chunks):
            data = [
                # - in some cases we need to convert time index to primitive type
                _time_cast_function(selected_table[k].chunk(n)).to_numpy() if k == _time_field_idx else selected_table[k].chunk(n).to_numpy()
                for k in range(selected_table.num_columns)]
            self._processor.process_data(data)
        return self._processor.get_result()
            


import numpy as np
cimport numpy as np


cdef np.ndarray nans(int dims)


cdef class Indexed:
    cdef public list values
    cdef public float max_series_length
    cdef unsigned short _is_empty


cdef class TimeSeries:
    cdef public long long timeframe
    cdef public Indexed times
    cdef public Indexed values
    cdef float max_series_length
    cdef unsigned short _is_new_item
    cdef public str name
    cdef dict indicators        # it's used for indicators caching
    cdef list calculation_order # calculation order as list: [ (input_id, indicator_obj, indicator_id) ]

    cdef _update_indicators(TimeSeries self, long long time, object value, short new_item_started)


cdef class Indicator(TimeSeries):
    cdef TimeSeries series
    cdef TimeSeries parent


cdef class RollingSum:
    """
    Rolling fast summator
    """
    cdef unsigned int period
    cdef np.ndarray __s
    cdef unsigned int __i
    cdef double rsum
    cdef unsigned short is_init_stage 

    cpdef double update(RollingSum self, double value, short new_item_started)


cdef class Bar:
    cdef public long long time
    cdef public double open
    cdef public double high
    cdef public double low
    cdef public double close
    cdef public double volume

    cpdef Bar update(Bar self, double price, double volume)

    cpdef dict to_dict(Bar self, unsigned short skip_time=*)


cdef class OHLCV(TimeSeries):
    cdef public TimeSeries open
    cdef public TimeSeries high
    cdef public TimeSeries low
    cdef public TimeSeries close
    cdef public TimeSeries volume

    cpdef short update(OHLCV self, long long time, double price, double volume=*)

    cpdef _update_indicators(OHLCV self, long long time, object value, short new_item_started)

    cpdef object append_data(
        OHLCV self, 
        np.ndarray times, 
        np.ndarray opens,
        np.ndarray highs, 
        np.ndarray lows,
        np.ndarray closes, 
        np.ndarray volumes
    )


cdef class Trade:
    cdef public long long time
    cdef public double price
    cdef public double size
    cdef public short taker


cdef class Quote:
    cdef public long long time
    cdef public bid
    cdef public ask
    cdef public bid_size
    cdef public ask_size

    cpdef double mid_price(Quote self)



import glob, re
import json, os, dataclasses
from datetime import datetime
from typing import Dict, List, Optional

from qubx.core.basics import Instrument, FuturesInfo
from qubx.utils.marketdata.binance import get_binance_symbol_info_for_type
from qubx import logger
from qubx.utils.misc import makedirs, get_local_qubx_folder


class _InstrumentEncoder(json.JSONEncoder):
    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return {k:v for k,v in dataclasses.asdict(obj).items() if not k.startswith('_')}
        if isinstance(obj, (datetime)):
            return obj.isoformat()
        return super().default(obj)


class _InstrumentDecoder(json.JSONDecoder):
    def _preprocess(d, ks):
        fi = d.get(ks)
        if fi:
            fi['delivery_date'] = datetime.strptime(fi.get('delivery_date', '5000-01-01T00:00:00'),'%Y-%m-%dT%H:%M:%S')
            fi['onboard_date'] = datetime.strptime(fi.get('onboard_date', '1970-01-01T00:00:00'),'%Y-%m-%dT%H:%M:%S')
        return d | {ks: FuturesInfo(**fi) if fi else None}

    def decode(self, json_string):
        obj = super(_InstrumentDecoder, self).decode(json_string)
        if isinstance(obj, dict):
            return Instrument(**_InstrumentDecoder._preprocess(obj, 'futures_info'))
        elif isinstance(obj, list):
            return [Instrument(**_InstrumentDecoder._preprocess(o, 'futures_info')) for o in obj]
        return obj


class InstrumentsLookup:
    lookup: Dict[str, Instrument]
    path: str

    def __init__(self, path: str=makedirs(get_local_qubx_folder(), 'instruments')) -> None:
        self.path = path
        if not self.load():
            self.refresh()
        self.load()

    def load(self) -> bool:
        self.lookup = {}
        data_exists = False
        for fs in glob.glob(self.path + '/*.json'):
            try:
                with open(fs, 'r') as f:
                    instrs = json.load(f, cls=_InstrumentDecoder)
                    for i in instrs:  
                        self.lookup[f"{i.exchange}:{i.symbol}"] = i
                    data_exists = True
            except Exception as ex:
                logger.warning(ex)

        return data_exists

    def find(self, exchange: str, base: str, quote: str) -> Optional[Instrument]:
        for i in self.lookup.values():
            if i.exchange == exchange and (
                (i.base == base and i.quote == quote) or (i.base == quote and i.quote == base)
            ):
                return i
        return None
    
    def find_aux_instrument_for(self, instrument: Instrument, base_currency: str) -> Optional[Instrument]:
        """
        Tries to find aux instrument (for conversions to funded currency)
        for example: 
            ETHBTC -> BTCUSDT for base_currency USDT
            EURGBP -> GBPUSD for base_currency USD
            ...
        """
        base_currency = base_currency.upper()
        if instrument.quote != base_currency and instrument._aux_instrument is None:
            return self.find(instrument.exchange, instrument.quote, base_currency)
        return instrument._aux_instrument

    def __getitem__(self, spath: str) -> List[Instrument]:
        res = []
        c = re.compile(spath)
        for k, v in self.lookup.items():
            if re.match(c, k):
                res.append(v)
        return res
    
    def refresh(self):
        for mn in dir(self):
            if mn.startswith('_update_'):
                getattr(self, mn)(self.path)

    def _update_kraken(self, path: str):
        # TODO
        pass

    def _update_dukas(self, path: str):
        instruments = [
            Instrument('EURUSD', 'FX', 'DUKAS', 'EUR', 'USD', 'USD', 0.00001, 1, 1000),
            Instrument('GBPUSD', 'FX', 'DUKAS', 'GBP', 'USD', 'USD', 0.00001, 1, 1000),
            Instrument('USDJPY', 'FX', 'DUKAS', 'USD', 'JPY', 'USD', 0.001,   1, 1000),
            Instrument('USDCAD', 'FX', 'DUKAS', 'USD', 'CAD', 'USD', 0.00001, 1, 1000),
            Instrument('AUDUSD', 'FX', 'DUKAS', 'AUD', 'USD', 'USD', 0.00001, 1, 1000),
            Instrument('USDPLN', 'FX', 'DUKAS', 'USD', 'PLN', 'USD', 0.00001, 1, 1000),
            Instrument('EURGBP', 'FX', 'DUKAS', 'EUR', 'GBP', 'USD', 0.00001, 1, 1000),
            # TODO: addd all or find how to get it from site
        ]
        logger.info(f'Updates {len(instruments)} for DUKASCOPY')
        with open(os.path.join(path, f'dukas.json'), 'w') as f:
            json.dump(instruments, f, cls=_InstrumentEncoder)

    def _update_binance(self, path: str):
        infos = get_binance_symbol_info_for_type(['UM', 'CM', 'SPOT'])
        for exchange, info in infos.items():
            instruments = []
            for s in info['symbols']:
                tick_size, size_step = None, None 
                for i in s['filters']:
                    if i['filterType'] == 'PRICE_FILTER':
                        tick_size = float(i['tickSize'])
                    if i['filterType'] == 'LOT_SIZE':
                        size_step = float(i['stepSize'])

                fut_info = None
                if 'contractType' in s:
                    fut_info = FuturesInfo( 
                        s.get('contractType', 'UNKNOWN'),
                        datetime.fromtimestamp(s.get('deliveryDate', 0)/1000.0),
                        datetime.fromtimestamp(s.get('onboardDate', 0)/1000.0),
                        float(s.get('contractSize', 1)), 
                        float(s.get('maintMarginPercent', 0)), 
                        float(s.get('requiredMarginPercent', 0)), 
                        float(s.get('liquidationFee', 0)),
                    )

                instruments.append(
                    Instrument(
                        s['symbol'], 'CRYPTO', exchange.upper(), s['baseAsset'], s['quoteAsset'], s.get('marginAsset', None), 
                        tick_size, size_step, 
                        min_size=size_step,    # TODO: not sure about minimal position for Binance
                        futures_info=fut_info
                ))

            logger.info(f'Loaded {len(instruments)} for {exchange}')

            with open(os.path.join(path, f'{exchange}.json'), 'w') as f:
                json.dump(instruments, f, cls=_InstrumentEncoder)
  


""""
Here stuff we want to have in every Jupyter notebook after calling %qube magic
"""
import importlib_metadata

import qubx
from qubx.utils import runtime_env
from qubx.utils.misc import add_project_to_system_path


def np_fmt_short():
    # default np output is 75 columns so extend it a bit and suppress scientific fmt for small floats
    np.set_printoptions(linewidth=240, suppress=True)


def np_fmt_reset():
    # reset default np printing options
    np.set_printoptions(edgeitems=3, infstr='inf', linewidth=75, nanstr='nan', precision=8,
                        suppress=False, threshold=1000, formatter=None)


if runtime_env() in ['notebook', 'shell']:

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # -- all imports below will appear in notebook after calling %%alphalab magic ---
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # - - - - Common stuff - - - -
    import numpy as np
    import pandas as pd
    from datetime import time, timedelta
    from tqdm.auto import tqdm

    # - - - - TA stuff and indicators - - - -
    # - - - - Portfolio analysis - - - -
    # - - - - Simulator stuff - - - -
    # - - - - Learn stuff - - - -
    # - - - - Charting stuff - - - -
    from matplotlib import pyplot as plt
    from qubx.utils.charting.mpl_helpers import fig, subplot, sbp
    # - - - - Utils - - - -

    # - setup short numpy output format
    np_fmt_short()
    
    # - add project home to system path
    add_project_to_system_path()

    # - check current version
    try: 
        version = importlib_metadata.version('qube2')
    except:
        version = 'Dev'

    # some new logo
    if not hasattr(qubx.QubxMagics, '__already_initialized__'):
        from qubx.utils.misc import (green, yellow, cyan, magenta, white, blue, red)

        print(
        f"""
                   {red("╻")}
   {green("┏┓      ╻     ")}  {red("┃")}  {yellow("┏┓")}       {cyan("Quantitative Backtesting Environment")} 
   {green("┃┃  ┓┏  ┣┓  ┏┓")}  {red("┃")}  {yellow("┏┛")}       
   {green("┗┻  ┗┻  ┗┛  ┗ ")}  {red("┃")}  {yellow("┗━")}       (c) 2024,  ver. {magenta(version.rstrip())}
                   {red("╹")}       
"""
        )
        qubx.QubxMagics.__already_initialized__ = True




from qubx.utils import set_mpl_theme, runtime_env, install_pyx_recompiler_for_dev
install_pyx_recompiler_for_dev()

from loguru import logger
import os, sys, stackprinter
from qubx.core.lookups import InstrumentsLookup


def formatter(record):
    end = record["extra"].get("end", "\n")
    fmt = "<lvl>{message}</lvl>%s" % end
    if record["level"].name in {"WARNING", "SNAKY"}:
        fmt = "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - %s" % fmt

    prefix = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> [ <level>%s</level> ] " % record["level"].icon

    if record["exception"] is not None:
        # stackprinter.set_excepthook(style='darkbg2')
        record["extra"]["stack"] = stackprinter.format(record["exception"], style="darkbg")
        fmt += "\n{extra[stack]}\n"

    if record["level"].name in {"TEXT"}:
        prefix = ""

    return prefix + fmt


config = {
    "handlers": [ {"sink": sys.stdout, "format": "{time} - {message}"}, ],
    "extra": {"user": "someone"},
}


logger.configure(**config)
logger.remove(None)
logger.add(sys.stdout, format=formatter, colorize=True)
logger = logger.opt(colors=True)

lookup = InstrumentsLookup()

# registering magic for jupyter notebook
if runtime_env() in ['notebook', 'shell']:
    from IPython.core.magic import (Magics, magics_class, line_magic, line_cell_magic)
    from IPython import get_ipython

    @magics_class
    class QubxMagics(Magics):
        # process data manager
        __manager = None

        @line_magic
        def qubxd(self, line: str):
            self.qubx_setup('dark')

        @line_magic
        def qubxl(self, line: str):
            self.qubx_setup('light')

        @line_magic
        def qubx_setup(self, line: str):
            """
            QUBE framework initialization
            """
            import os

            tpl_path = os.path.join(os.path.dirname(__file__), "_nb_magic.py")
            # print("TPL:", tpl_path)
            with open(tpl_path, 'r', encoding="utf8") as myfile:
                s = myfile.read()

            exec(s, self.shell.user_ns)

            # setup more funcy mpl theme instead of ugly default
            if line:
                if 'dark' in line.lower():
                    set_mpl_theme('dark')

                elif 'light' in line.lower():
                    set_mpl_theme('light')

            # install additional plotly helpers
            # from qube.charting.plot_helpers import install_plotly_helpers
            # install_plotly_helpers()

        def _get_manager(self):
            if self.__manager is None:
                import multiprocessing as m
                self.__manager = m.Manager()
            return self.__manager

        @line_cell_magic
        def proc(self, line, cell=None):
            """
            Run cell in separate process

            >>> %%proc x, y as MyProc1
            >>> x.set('Hello')
            >>> y.set([1,2,3,4])
            
            """
            import multiprocessing as m
            import time, re

            # create ext args
            name = None
            if line:
                # check if custom process name was provided
                if ' as ' in line:
                    line, name = line.split('as')
                    if not name.isspace():
                        name = name.strip()
                    else:
                        print('>>> Process name must be specified afer "as" keyword !')
                        return

                ipy = get_ipython()
                for a in [x for x in re.split('[\ ,;]', line.strip()) if x]:
                    ipy.push({a: self._get_manager().Value(None, None)})

            # code to run
            lines = '\n'.join(['    %s' % x for x in cell.split('\n')])

            def fn():
                result = get_ipython().run_cell(lines)

                # send errors to parent
                if result.error_before_exec:
                    raise result.error_before_exec

                if result.error_in_exec:
                    raise result.error_in_exec

            t_start = str(time.time()).replace('.', '_')
            f_id = f'proc_{t_start}' if name is None else name
            if self._is_task_name_already_used(f_id):
                f_id = f"{f_id}_{t_start}"

            task = m.Process(target=fn, name=f_id)
            task.start()
            print(' -> Task %s is started' % f_id)

        def _is_task_name_already_used(self, name):
            import multiprocessing as m
            for p in m.active_children():
                if p.name == name:
                    return True
            return False

        @line_magic
        def list_proc(self, line):
            import multiprocessing as m
            for p in m.active_children():
                print(p.name)

        @line_magic
        def kill_proc(self, line):
            import multiprocessing as m
            for p in m.active_children():
                if line and p.name.startswith(line):
                    p.terminate()


    # - registering magic here
    get_ipython().register_magics(QubxMagics)



from typing import Dict, Union
from datetime import timedelta
import pandas as pd
import numpy as np

from numpy.lib.stride_tricks import as_strided as stride

from qubx.utils.misc import Struct


def drop_duplicated_indexes(df, keep='first'):
    """
    Drops duplicated indexes in dataframe/series
    Keeps either first or last occurence (parameter keep)
    """
    return df[~df.index.duplicated(keep=keep)]


def process_duplicated_indexes(data: Union[pd.DataFrame, pd.Series], ns=1) -> Union[pd.DataFrame, pd.Series]:
    """
    Finds duplicated indexes in frame/series and add shift (in nS) to every repeating one
    :param data: time indexed dataframe/series
    :param ns: shift constant in nanosec
    :return: return dataframe with all no duplicated rows (each duplicate has own unique index)
    """
    values = data.index.duplicated(keep='first').astype(float)
    values[values == 0] = np.NaN

    missings = np.isnan(values)
    cumsum = np.cumsum(~missings)
    diff = np.diff(np.concatenate(([0.], cumsum[missings])))
    values[missings] = -diff

    # set new index (1 ms)
    data.index = data.index.values + np.cumsum(values).astype(np.timedelta64) * ns
    return data


def scols(*xs, keys=None, names=None, keep='all') -> pd.DataFrame:
    """
    Concat dataframes/series from xs into single dataframe by axis 1
    :param keys: keys of new dataframe (see pd.concat's keys parameter)
    :param names: new column names or dict with replacements
    :return: combined dataframe
    
    Example
    -------
    >>>  scols(
            pd.DataFrame([1,2,3,4,-4], list('abcud')),
            pd.DataFrame([111,21,31,14], list('xyzu')), 
            pd.DataFrame([11,21,31,124], list('ertu')), 
            pd.DataFrame([11,21,31,14], list('WERT')), 
            names=['x', 'y', 'z', 'w'])
    """
    r = pd.concat((xs), axis=1, keys=keys)
    if names:
        if isinstance(names, (list, tuple)):
            if len(names) == len(r.columns):
                r.columns = names
            else:
                raise ValueError(
                    f"if 'names' contains new column names it must have same length as resulting df ({len(r.columns)})")
        elif isinstance(names, dict):
            r = r.rename(columns=names)
    return r


def srows(*xs, keep='all', sort=True) -> Union[pd.DataFrame, pd.Series]:
    """
    Concat dataframes/series from xs into single dataframe by axis 0
    :param sort: if true it sorts resulting dataframe by index (default)
    :param keep: how to deal with duplicated indexes. 
                 If set to 'all' it doesn't do anything (default). Otherwise keeps first or last occurences
    :return: combined dataframe
    
    Example
    -------
    >>>  srows(
            pd.DataFrame([1,2,3,4,-4], list('abcud')),
            pd.DataFrame([111,21,31,14], list('xyzu')), 
            pd.DataFrame([11,21,31,124], list('ertu')), 
            pd.DataFrame([11,21,31,14], list('WERT')), 
            sort=True, keep='last')
    """
    r = pd.concat((xs), axis=0)
    r = r.sort_index() if sort else r
    if keep != 'all':
        r = drop_duplicated_indexes(r, keep=keep)
    return r


def retain_columns_and_join(data: dict, columns) -> pd.DataFrame:
    """
    Retains given columns from every value of data dictionary and concatenate them into single data frame

    from qube.datasource import DataSource
    from qube.analysis.tools import retain_columns_and_join

    ds = DataSource('yahoo::daily')
    data = ds.load_data(['aapl', 'msft', 'spy'], '2000-01-01', 'now')

    closes = retain_columns_and_join(data, 'close')
    hi_lo = retain_columns_and_join(data, ['high', 'low'])

    :param data: dictionary with dataframes
    :param columns: columns names need to be retained
    :return: data frame
    """
    if not isinstance(data, dict):
        raise ValueError('Data must be passed as dictionary')

    return pd.concat([data[k][columns] for k in data.keys()], axis=1, keys=data.keys())


def continuous_periods(xs, cond) -> Struct:
    """
    Detect continues periods on series xs based on condition cond
    """
    df = scols(xs, cond, keys=['_XS_', 'sig'])
    df['block'] = (df.sig.shift(1) != df.sig).astype(int).cumsum()
    idx_col_name = xs.index.name

    blk = df[df.sig].reset_index().groupby('block')[idx_col_name].apply(np.array)
    starts = blk.apply(lambda x: x[0])
    ends = blk.apply(lambda x: x[-1])
    se_info = scols(starts, ends, keys=['start', 'end'])
    return Struct(blocks=blk.reset_index(drop=True), periods=se_info)


def roll(df: pd.DataFrame, w: int, **kwargs):
    """
    Rolling window on dataframe using multiple columns
    
    >>> roll(pd.DataFrame(np.random.randn(10,3), index=list('ABCDEFGHIJ')), 3).apply(print)
    
    or alternatively 
    
    >>> pd.DataFrame(np.random.randn(10,3), index=list('ABCDEFGHIJ')).pipe(roll, 3).apply(lambda x: print(x[2]))
    
    :param df: pandas DataFrame
    :param w: window size (only integers)
    :return: rolling window
    """
    if w > len(df):
        raise ValueError("Window size exceeds number of rows !")

    v = df.values
    d0, d1 = v.shape
    s0, s1 = v.strides
    a = stride(v, (d0 - (w - 1), w, d1), (s0, s0, s1))
    rolled_df = pd.concat({
        row: pd.DataFrame(values, columns=df.columns)
        for row, values in zip(df.index, a)
    })

    return rolled_df.groupby(level=0, **kwargs)


def dict_to_frame(x: dict, index_type=None, orient='index', columns=None, column_types=dict()) -> pd.DataFrame:
    """
    Utility for convert dictionary to indexed DataFrame
    It's possible to pass columns names and type of index
    """
    y = pd.DataFrame().from_dict(x, orient=orient)
    if index_type:
        if index_type in ['ns', 'nano']:
            index_type = 'M8[ns]'
        y.index = y.index.astype(index_type)

    # rename if needed
    if columns:
        columns = [columns] if not isinstance(columns, (list, tuple, set)) else columns
        if len(columns) == len(y.columns):
            y.rename(columns=dict(zip(y.columns, columns)), inplace=True)
        else:
            raise ValueError('dict_to_frame> columns argument must contain %d elements' % len(y.columns))

    # if additional conversion is required
    if column_types:
        _existing_cols_conversion = {c: v for c, v in column_types.items() if c in y.columns}
        y = y.astype(_existing_cols_conversion)

    return y


def select_column_and_join(data: Dict[str, pd.DataFrame], column: str) -> pd.DataFrame:
    """
    Select given column from every value of data dictionary and concatenate them into single data frame

    from qube.datasource import DataSource
    from qube.analysis.tools import retain_columns_and_join

    ds = DataSource('yahoo::daily')
    data = ds.load_data(['aapl', 'msft', 'spy'], '2000-01-01', 'now')

    closes = select_column_and_join(data, 'close')
    hi_lo = select_column_and_join(data, ['high', 'low'])

    :param data: dictionary with dataframes
    :param columns: column name need to be selected
    :return: pandas data frame
    """
    if not isinstance(data, dict):
        raise ValueError('Data must be passed as dictionary of pandas dataframes')

    return pd.concat([data[k][column] for k in data.keys()], axis=1, keys=data.keys())



"""
 # All interfaces related to strategy etc
"""
from typing import Callable, Dict, List, Optional, Union
import numpy as np
from dataclasses import dataclass
from qubx.core.basics import ZERO_COSTS, Instrument, Position, Signal, TransactionCostsCalculator, dt_64

E_TIMER = 1
E_QUOTE = 2
E_TRADE = 3
E_OPENBOOK = 4
E_HIST_DATA_READY = 100
E_HIST_DATA_ERROR = -100
@dataclass
class Event:
    time: dt_64
    type: int               # ??
    instrument: Instrument

DataListener = Callable[[Instrument, int], None]
ExchListener = Callable[[Instrument, int], None]

class DataProvider:
    def add_data_listener(self, listener: DataListener):
        pass

    def request_historical_data(self, 
                                instruments: List[Instrument], 
                                timeframe: str,
                                start: Union[str, int, dt_64], 
                                stop: Union[str, int, dt_64]):
        pass


class ExchangeServiceProvider:
    def add_exchange_listener(self, listener: ExchListener):
        pass

    def get_position(self, instrument: Instrument) -> Position:
        pass

    def get_tcc(self, instrument: Instrument) -> TransactionCostsCalculator:
        return ZERO_COSTS

    def time(self) -> dt_64:
        """
        Returns current time
        """
        pass
    

class IStrategy:
    ctx: 'TradingContext'

    def on_init(self):
        pass

    def process_event(self, time: dt_64, event: Event) -> Optional[List[Signal]]:
        return None

        
class TradingContext:
    strategy: IStrategy 
    exchange: ExchangeServiceProvider
    data: DataProvider
    instruments: List[Instrument]
    positions: Dict[str, Position]

    def __init__(
            self, 
            strategy: IStrategy, 
            exchange: ExchangeServiceProvider, 
            data: DataProvider,
            instruments: List[Instrument]
        ) -> None:
        self.strategy = strategy
        self.exchange = exchange
        self.data = data

        self.instruments = []
        self.positions = {}
        for instr in instruments:
            # process instruments - need to find convertors etc
            # . . . . 
            self.instruments.append(instr)
            p = exchange.get_position(instr)
            self.positions[instr.symbol] = Position(instr) 




from qubx.utils import convert_tf_str_td64
import numpy as np
cimport numpy as np


NS = 1_000_000_000 

cpdef recognize_time(time):
    return np.datetime64(time, 'ns') if isinstance(time, str) else np.datetime64(time, 'ms')


cpdef str time_to_str(long long t, str units = 'ns'):
    return str(np.datetime64(t, units)) #.isoformat()


cpdef str time_delta_to_str(long long d):
    """
    Convert timedelta object to pretty print format

    :param d:
    :return:
    """
    days, seconds = divmod(d, 86400*NS)
    hours, seconds = divmod(seconds, 3600*NS)
    minutes, seconds = divmod(seconds, 60*NS)
    seconds, rem  = divmod(seconds, NS)
    r = ''
    if days > 0:
        r += '%dD' % days
    if hours > 0:
        r += '%dH' % hours
    if minutes > 0:
        r += '%dMin' % minutes
    if seconds > 0:
        r += '%dS' % seconds
    if rem > 0:
        r += '%dmS' % (rem // 1000000)
    return r


cpdef recognize_timeframe(timeframe):
    tf = timeframe
    if isinstance(timeframe, str):
        tf = np.int64(convert_tf_str_td64(timeframe).item().total_seconds() * NS)

    elif isinstance(timeframe, (int, float)) and timeframe >= 0:
        tf = timeframe

    elif isinstance(timeframe, np.timedelta64):
        tf = np.int64(timeframe.item().total_seconds() * NS) 

    else:
        raise ValueError('Unknown timeframe type !')
    return tf


from .time import convert_seconds_to_str, convert_tf_str_td64, floor_t64, time_to_str, infer_series_frequency
from .misc import install_pyx_recompiler_for_dev, runtime_env, version, Struct
from .charting.mpl_helpers import set_mpl_theme, fig, sbp, vline, hline, ellips, set_mpl_theme


import glob, os
from collections import OrderedDict, namedtuple
from os.path import basename, exists, dirname, join, expanduser
from typing import Optional, Union
from pathlib import Path


def version() -> str:
    # - check current version
    version = 'Dev'
    try: 
        import importlib_metadata
        version = importlib_metadata.version('qube2')
    except:
        pass

    return version


from ._pyxreloader import pyx_install_loader

def install_pyx_recompiler_for_dev():
    if version().lower() == 'dev':
        print(f" >  [{green('dev')}] {red('installed cython rebuilding hook')}")
        pyx_install_loader(['qubx.core', 'qubx.ta', 'qubx.data', 'qubx.strategies'])


def runtime_env():
    """
    Check what environment this script is being run under
    :return: environment name, possible values:
             - 'notebook' jupyter notebook
             - 'shell' any interactive shell (ipython, PyCharm's console etc)
             - 'python' standard python interpreter
             - 'unknown' can't recognize environment
    """
    try:
        from IPython import get_ipython
        shell = get_ipython().__class__.__name__

        if shell == 'ZMQInteractiveShell':  # Jupyter notebook or qtconsole
            return 'notebook'
        elif shell.endswith('TerminalInteractiveShell'):  # Terminal running IPython
            return 'shell'
        else:
            return 'unknown'  # Other type (?)
    except (NameError, ImportError):
        return 'python'  # Probably standard Python interpreter

_QUBX_FLDR = None

def get_local_qubx_folder() -> str:
    global _QUBX_FLDR

    if _QUBX_FLDR is None:
        _QUBX_FLDR = makedirs(os.getenv('QUBXSTORAGE', os.path.expanduser('~/.qubx')))

    return _QUBX_FLDR


def add_project_to_system_path(project_folder:str = '~/projects'):
    """
    Add path to projects folder to system python path to be able importing any modules from project
    from test.Models.handy_utils import some_module
    """
    import sys
    from os.path import expanduser, relpath
    from pathlib import Path
    
    # we want to track folders with these files as separate paths
    toml = Path('pyproject.toml')
    src = Path('src')
    
    try:
        prj = Path(relpath(expanduser(project_folder)))
    except ValueError as e:
        # This error can occur on Windows if user folder and python file are on different drives
        print(f"Qube> Error during get path to projects folder:\n{e}")
    else:
        insert_path_iff = lambda p: sys.path.insert(0, p.as_posix()) if p.as_posix() not in sys.path else None
        if prj.exists():
            insert_path_iff(prj)
            
            for di in prj.iterdir():
                _src = di / src
                if (di / toml).exists():
                    # when we have src/
                    if _src.exists() and _src.is_dir():
                        insert_path_iff(_src)
                    else:
                        insert_path_iff(di)
        else:
            print(f'Qube> Cant find {project_folder} folder for adding to python path !')


def is_localhost(host):
    return host.lower() == 'localhost' or host == '127.0.0.1'


def __wrap_with_color(code):
    def inner(text, bold=False):
        c = code
        if bold:
            c = "1;%s" % c
        return "\033[%sm%s\033[0m" % (c, text)

    return inner


red, green, yellow, blue, magenta, cyan, white = (
    __wrap_with_color('31'),
    __wrap_with_color('32'),
    __wrap_with_color('33'),
    __wrap_with_color('34'),
    __wrap_with_color('35'),
    __wrap_with_color('36'),
    __wrap_with_color('37'),
)


class Struct:
    """
    Dynamic structure (similar to matlab's struct it allows to add new properties dynamically)

    >>> a = Struct(x=1, y=2)
    >>> a.z = 'Hello'
    >>> print(a)

    Struct(x=1, y=2, z='Hello')
    
    >>> Struct(a=234, b=Struct(c=222)).to_dict()
    
    {'a': 234, 'b': {'c': 222}}

    >>> Struct({'a': 555}, a=123, b=Struct(c=222)).to_dict()

    {'a': 123, 'b': {'c': 222}}
    """

    def __init__(self, *args, **kwargs):
        _odw = OrderedDict(**kwargs)
        if args:
            if isinstance(args[0], dict):
                _odw = OrderedDict(Struct.dict2struct(args[0]).to_dict()) | _odw
            elif isinstance(args[0], Struct):
                _odw = args[0].to_dict() | _odw
        self.__initialize(_odw.keys(), _odw.values())

    def __initialize(self, fields, values):
        self._fields = list(fields)
        self._meta = namedtuple('Struct', ' '.join(fields))
        self._inst = self._meta(*values)

    def fields(self) -> list:
        return self._fields

    def __getitem__(self, idx: int):
        return getattr(self._inst, self._fields[idx])

    def __getattr__(self, k):
        return getattr(self._inst, k)

    def __or__(self, other: Union[dict, 'Struct']):
        if isinstance(other, dict):
            other = Struct.dict2struct(other)
        elif not isinstance(other, Struct):
            raise ValueError(f"Can't union with object of {type(other)} type ")
        for f in other.fields():
            self.__setattr__(f, other.__getattr__(f))
        return self

    def __dir__(self):
        return self._fields

    def __repr__(self):
        return self._inst.__repr__()

    def __setattr__(self, k, v):
        if k not in ['_inst', '_meta', '_fields']:
            new_vals = {**self._inst._asdict(), **{k: v}}
            self.__initialize(new_vals.keys(), new_vals.values())
        else:
            super().__setattr__(k, v)

    def __getstate__(self):
        return self._inst._asdict()

    def __setstate__(self, state):
        self.__init__(**state)

    def __ms2d(self, m) -> dict:
        r = {}
        for f in m._fields:
            v = m.__getattr__(f)
            r[f] = self.__ms2d(v) if isinstance(v, Struct) else v
        return r

    def to_dict(self) -> dict:
        """
        Return this structure as dictionary
        """
        return self.__ms2d(self)

    def copy(self) -> 'Struct':
        """
        Returns copy of this structure
        """
        return Struct(self.to_dict())

    @staticmethod
    def dict2struct(d: dict) -> 'Struct':
        """
        Convert dictionary to structure
        >>> s = dict2struct({'f_1_0': 1, 'z': {'x': 1, 'y': 2}})
        >>> print(s.z.x)
        1
        """
        m = Struct()
        for k, v in d.items():
            # skip if key is not valid identifier
            if not k.isidentifier():
                print(f"Struct> {k} doesn't look like as identifier - skip it")
                continue
            if isinstance(v, dict):
                v = Struct.dict2struct(v)
            m.__setattr__(k, v)
        return m


def makedirs(path: str, *args) -> str:
    path = os.path.expanduser(os.path.join(*[path, *args]))
    if not exists(path):
        os.makedirs(path)
    return path